%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Blocking
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(blocking,
	[blocking_clause_set/2
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').


blocking_clause_set(CS, Res) :-
        scan_function_symbols(CS, FSig),
        ( nb_getval(rangerestriction_option, off) ->
            make_f_term_predicate(FSig, ExtraClauses)
        ; %% without range restriction we don't have a dom predicate
          %% and no means to axiomatize the subterm relationship.
          %% We speculate equalities possibly between arbitrary
          %% terms then, however confined to the specific blocking
          %% option set.
          nb_getval(unrestrictedblocking_option, true) ->
          %% same if unrestricted blocking is set
            ExtraClauses = []
        ; %% otherwise ancestor-restricted blocking
            subterm_clauses(FSig, ExtraClauses)
        ),
 	scan_predicate_symbols(CS, PSig),
	( nb_getval(blocking_option, dom) ->
            ( nb_getval(rangerestriction_option, off) -> 
                msg('WARNING: blocking is \'dom\' but range restriction is off.',[]),
                msg('Assuming e-krh\'s built-in dom-predicate.',[]),
                speculate_equality_clauses(['$krh_dom'/1], EQClauses)
            ; speculate_equality_clauses([dom/1], EQClauses) 
            )
        ; nb_getval(blocking_option, unary) ->
             findall(P/1, member(P/1, PSig), UPSig),
             delete(UPSig, dom/1, UPSig1),
            speculate_equality_clauses(UPSig1, EQClauses)
        ; %% must be n-ary
            %% we don't do it for dom:
             delete(PSig, dom/1, PSig1),
             %% and we don't do it for equality:
             delete(PSig1, (=)/2, PSig2),
             % delete(PSig2, my_equal/2, PSig3),
             speculate_equality_clauses(PSig2, EQClauses)
        ),
	append([CS,ExtraClauses,EQClauses], Res).


%% Proper subterm
subterm_clauses(Sig, Res) :-
	%% Sig is a list of function symbols plus arities (F/N)
	maplist(subterm_clauses_1, Sig, HRes1),
        append(HRes1, HRes1Flat),
        maplist(subterm_clauses_2, Sig, HRes2),
        append(HRes2, HRes2Flat),
	append(HRes1Flat, HRes2Flat, Res).

subterm_clauses_1(F/N, Res) :-
	%% See example below to make sense of this
	make_N_variables(N, NVars),
	FNVars =.. [F|NVars],
	findall(([subterm(X,FNVars)] :- 
%                        [dom(X), dom(FNVars)]
                        [dom(FNVars)]
%                        []
                ),
	        member(X, NVars),
		Res).

subterm_clauses_2(F/N, Res) :-
	%% See example below to make sense of this
	make_N_variables(N, NVars),
	FNVars =.. [F|NVars],
	findall(([subterm(X,FNVars)] :- 
                        [dom(FNVars), subterm(X,Y)]),
	        member(Y, NVars),
		Res).

% subterm_clauses(Sig, [([subterm(X,X)] :- [dom(X)])|Res]) :-
% 	%% Sig is a list of function symbols plus arities (F/N)
% 	maplist(subterm_clauses_, Sig, HRes),
% 	%% Have a list of lists now, flatten it
% 	append(HRes, Res).

% subterm_clauses_(F/N, Res) :-
% 	%% See example below to make sense of this
% 	make_N_variables(N, NVars),
% 	FNVars =.. [F|NVars],
% 	findall(([subterm(X,FNVars)] :- 
%                         [subterm(X,Y), dom(X), dom(FNVars)]),
% 	        member(Y, NVars),
% 		Res).

make_f_term_predicate(FSig, Res) :-
        findall(([f_term(T)] :- []),
                ( member(F/N, FSig), N > 0,
                  make_N_variables(N, NXVars),
                  T =.. [F|NXVars] ),
                Res).

make_constant_predicate(FSig, Res) :-
        findall(([constant(T)] :- []),
                member(T/0, FSig),
                Res).

speculate_equality_clauses(Sig, [% ([synequal(X,X)] :- []),
                                 ([] :- [nonequal(X,X)]), 
	                         ([] :- [(Y=X), nonequal(Y,X)])|
			         Res]) :-
	%% Sig is a list of function symbols plus arities (F/N)
	maplist(speculate_equality_clauses_, Sig, HRes),
	%% Have a list of lists now, flatten it
	append(HRes, Res).

speculate_equality_clauses_(F/N, Res) :-
	make_N_variables(N, NXVars),
	make_N_variables(N, NYVars),
	FNXVars =.. [F|NXVars],
	FNYVars =.. [F|NYVars],
        ( ( nb_getval(rangerestriction_option, off) ;
            nb_getval(unrestrictedblocking_option, true) ) ->
            %% don't confine to subterms
            findall(([Y=X, nonequal(Y,X)] :- [FNXVars, FNYVars %,
                                              %f_term(X) %, 
                                              %not_subsumed(synequal(X, Y))
                                             ]),
	        (member2(X, Y, NXVars, NYVars)),
		Res)
        ; findall(([Y=X, nonequal(Y,X)] :- [FNYVars, 
                                            %% constant(X),
                                            subterm(X,Y), 
                                            FNXVars]),
	        member2(X, Y, NXVars, NYVars),
		Res)).
	

/*
 * Example:

subterm(X,X) :- dom(X).

subterm(X,f(Y)) :- subterm(X,Y), dom(X), dom(f(Y)).

subterm(X,g(Y1,Y2)) :- subterm(X,Y1), dom(X), dom(g(Y1,Y2)).
subterm(X,g(Y1,Y2)) :- subterm(X,Y2), dom(X), dom(g(Y1,Y2)).

false :- Y=X, nonequal(Y,X).

Y1=X1 ; nonequal(Y1,X1) :- r(X1,X2), r(Y1,Y2), subterm(X1,Y1).
Y2=X2 ; nonequal(Y2,X2) :- r(X1,X2), r(Y1,Y2), subterm(X2,Y2).


 */
