%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Separation
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(rangerestriction, 
	[rr_clause_set/2,
	rr_clause/2
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').
%% :- use_module('../yarralumla/sorting').

rr_clause_set(CS, Res) :-
	%% (1) add the collapse constant to the domain:
 	scan_function_symbols(CS, FSig),
 	( member(AConstant/0, FSig) ; AConstant = the_collapse_constant ),
	Res1 = [([dom(AConstant)] :- [])],
	%% Ugly, but we need the value below:
	nb_setval(the_collapse_constant, AConstant),
	%% (2) Make the clauses that insert new domain elements for
	%% atoms derived. These clauses are based on the signature ...
 	scan_predicate_symbols(CS, CSPSig),
	%% ... but usually not for equality,
	%% because it will leave the axiom
	%% dom(X) :- X=Y.
	%% dom(Y) :- X=Y.
	%% which will be simplified by SPASS to 
	%% dom(X).
	%% And this is unusable for us.
	%% Instead, all head equations should have been renamed away when
	%% using a prover like SPASS:
	( nb_getval(headequationrenaming_option, true) ->
	    %% Thus, not for equality:
	    delete(CSPSig, (=)/2, CSPSigForDom1),
	    %% but for the new predicate symbol 'head_equation':
	    CSPSigForDomAxioms = [head_equation/2|CSPSigForDom1],
	    %% rename the (non-flat) head equations:
	    maplist(rename_head_equations_clause, CS, CS1H),
	    %% We leave the constants in-place and add domain axioms for them:
	    sublist(is_constant, FSig, CSig),
	    maplist(symbol_arity_symbol, CSig, Cs),
	    %% Fs is a list of constants, thus suitable:
	    domain_elements_axioms_from_terms(Cs, [], Res2),
	    %% Add the definition for the 'head_equation' predicate symbol:
	    CS1 = [([X=Y] :- [head_equation(X,Y)])|CS1H]
	; %% otherwise treat equations like any other predicate symbols:
	  CSPSigForDomAxioms = CSPSig,
	  %% and of course no renaming:
	  CS1 = CS,
	  Res2 = []
        ),
	%% Make the clauses that insert the domain elements for the
	%% heads that have become true:
 	domain_elements_axioms_from_sig(CSPSigForDomAxioms, Res3),
	%%
	%% (3) Make the clauses that insert remaining new domain elements.
	%% These clauses insert domain elements depending from the
	%% terms in the body:
	maplist(domain_elements_axioms_from_clause_body, CS1, DomElemAxioms4L),
	%% Have a list of lists, thus flatten ...
	append(DomElemAxioms4L, DomElemAxioms4),
	%% ... and range-restrict:
	maplist(rr_clause, DomElemAxioms4, Res4),
	%%
	%% (4) Range-restrict the current clause set:
	maplist(rr_clause, CS1, Res5),
	%% Add clauses to make with each term its subterms domain elements,
	%% too:
 	dom_subterm_clauses(FSig, Res6),
	%% Finally collect the grand result:
	append([Res1, Res2, Res3, Res4, Res5, Res6], Res).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rr_clause((H :- B), (H :- BDom)) :-
	term_variables(H, HVars),
	term_variables(B, BVars),
	%% For *extra* variables  ...
	subtractq(HVars, BVars, HDomVars), 
	%% and for variables occuring in body equations:
  	p_atoms(B, (=), BEquations),
  	term_variables(BEquations, BEquationsVars),
  	unionq(HDomVars, BEquationsVars, DomVars),
	make_doms(DomVars, DomVarsDom),
	append(B, DomVarsDom, BDom).

make_doms([],[]).
make_doms([X|R],[dom(X)|DomR]) :- make_doms(R,DomR).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% domain_elements_axioms_from_clause_body((_H :- B), Res)
%% Make clauses to insert domain elements depending from the
%% terms in the body:

domain_elements_axioms_from_clause_body((_H :- B), Res) :-
	%% Make the clauses to insert the terms 
	%% occuring in the body.
	maplist(domain_elements_axioms_from_body_lit, B, 
	             InsertDomElementsClauses1),
	%% Have a list of list clauses, thus flatten:
	append(InsertDomElementsClauses1, Res).


/*
 * Obsoleted by domain_elements_axioms_from_clause_body above
 * and renaming of head equations:
 *
domain_elements_axioms_from_clause((H :- B), Res) :-
	%% First, make the clauses to insert the terms 
	%% occuring in the body.
	maplist(domain_elements_axioms_from_body_lit, B, 
	             InsertDomElementsClauses1L),
	%% Have a list of list clauses, thus flatten:
	append(InsertDomElementsClauses1L, InsertDomElementsClauses1),
	%% Second, make the clauses to insert the terms 
	%% occuring in the head, BUT ONLY FOR EQUATIONS.
	%% For non-equations we do something different, see ALTERNATIVE above.
	%% Get all equations in the head:
	p_atoms(H, (=), HEqs), 
	%% Get all terms in the head:
	terms_args_union(HEqs, HEqsTerms),
	%% Restrict to the non-variable terms:
	sublist(nonvar, HEqsTerms, HEqsFunTerms),
	%% Make the clauses that insert the domain elements.
	%% Notice the guard, B, 
	%% the body of the clause.
	%% This guard is OK, because the
	%% new domain elements have to be made only if the rule "fires",
	%% i.e. when the body is satisfied.
  	domain_elements_axioms_from_terms(HEqsFunTerms, B, 
  	                           InsertDomElementsClauses2),
	append(InsertDomElementsClauses1, InsertDomElementsClauses2, Res).
*/

domain_elements_axioms_from_body_lit(A, Res) :-
	nb_getval(the_collapse_constant, AConstant),
	abstract_atom(A, AAbst, Subst), 
	findall(([dom(T)] :- [AAbst]),
	        ( member(X/T, Subst),
		  nonvar(T),
		  X = AConstant
	        ),
		Res).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% domain_elements_axioms_from_terms(Terms, Body, Res)
%% Terms is a list of terms.
%% Res is a list of clauses with
%% heads dom(T) and body Body, for each term T in Terms.

domain_elements_axioms_from_terms([], _Body, []).
domain_elements_axioms_from_terms([T|R], Body, [([dom(T)] :- Body)|RRes]) :-
	domain_elements_axioms_from_terms(R, Body, RRes).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

domain_elements_axioms_from_sig(PredSymbs, DomElAxioms) :-
	maplist(domain_elements_axioms_from_sig_1, PredSymbs, DomElAxiomsL),
	append(DomElAxiomsL, DomElAxioms).

domain_elements_axioms_from_sig_1(P/N, Res) :-
	make_N_variables(N, Vars),
	PBody =.. [P|Vars],
	domain_elements_axioms_from_sig_2(Vars, [PBody], Res).

domain_elements_axioms_from_sig_2([], _Body, []).
domain_elements_axioms_from_sig_2([X|Res], Body, [([dom(X)] :- Body)|RRes]) :-
	domain_elements_axioms_from_sig_2(Res, Body, RRes).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dom_subterm_clauses(Sig, Res) :-
	%% Sig is a list of function symbols plus arities (F/N)
	maplist(dom_subterm_clauses_1, Sig, HRes),
	%% Have a list of lists now, flatten it
	append(HRes, Res).

dom_subterm_clauses_1(F/N, Res) :-
	%% See example below to make sense of this
	make_N_variables(N, NXVars),
	FNXVars =.. [F|NXVars],
	findall(([dom(X)] :- [dom(FNXVars)]),
	        member(X, NXVars),
		Res).

/*
 * Example:

dom(X1) :- dom(g(X1,X2)).
dom(X2) :- dom(g(X1,X2)).

 */

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

abstract_atom(A, ARes, Subst) :-
	A =.. [P|Args],
	abstract_arg_list(Args, [], ArgsRes, _NewVarsSoFar, Subst),
	ARes =.. [P|ArgsRes].

%% abstract_arg_list(Args, VarsSoFar, ArgsRes, NewVarsSoFar, ArgsSubst):
abstract_arg_list([], VarsSoFar, [], VarsSoFar, []).
abstract_arg_list([X|R], VarsSoFar, [X|RAbst], NewVarsSoFar, RSubst) :-
	%% If X is a variable that has not occured before then it may remain:
	%% var(X),	\+ memberq(X, VarsSoFar), !,
	%% Alternative: all variables may remain:
	var(X),	
	%% Just go on with the rest, but remember X is present now:
	abstract_arg_list(R, [X|VarsSoFar], RAbst, NewVarsSoFar, RSubst).
abstract_arg_list([T|R], VarsSoFar, [X|RAbst], NewVarsSoFar, [(X/T)|RSubst]) :-
	%% In all other cases the term T has to be replaced by a fresh 
	%% variable, which we call X here.
	%% (No need to remember X, as it is fresh anyways)
	abstract_arg_list(R, VarsSoFar, RAbst, NewVarsSoFar, RSubst).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rename_head_equations_clause((H :- B), (HRes :- B)) :-
	rename_nonflat_equations(H, HRes).

rename_nonflat_equations([], []).
rename_nonflat_equations([(S=T)|R], [head_equation(S,T)|RRes]) :-
	%% The equation contains a non-flat term:
	member(U, [S,T]),
	compound(U), !,
	rename_nonflat_equations(R, RRes).
rename_nonflat_equations([A|R], [A|RRes]) :-
	rename_nonflat_equations(R, RRes).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_constant(_F/0).

symbol_arity_symbol(F/_, F).
