%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Separation
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(rangerestriction, 
	[rr_clause_set/2,
	rr_clause/2
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').
:- use_module('../yarralumla/sorting').

rr_clause_set(HCS, Res) :-
	%% Peliminary: add reflexivity:
	CS = [([X=X] :- [])|HCS],
	%% (1) Make the clauses that insert new domain elements for
	%% atoms derived. These clauses are based on the signature ...
 	scan_predicate_symbols(CS, CSPSig),
	%% ... but not for equality (see ALTERNATIVE below):
 	delete(CSPSig, (=)/2, CSPSigNE),
 	domain_elements_axioms_from_sig(CSPSigNE, Res1),
	%%
	%% (2) Make the clauses that insert remaining new domain elements.
	%% These clauses insert domain elements 
	%% (a) for the terms in the bodies, and
	%% (b) for the terms occuring in the equations in the heads:
	maplist(domain_elements_axioms_from_clause, CS, DomElemAxioms2L),
	%% Have a list of lists, thus flatten ...
	append(DomElemAxioms2L, DomElemAxioms2),
	%% ... and range-restrict:
	maplist(rr_clause, DomElemAxioms2, Res2),
	%%
	%% (3) Range-restrict the given clause set:
	maplist(rr_clause, CS, Res3),
	%%
	%% (4) Make the clauses
	%% to union the sorts of the variables of the clauses in CS:
	maplist(sort_clause_clauses, CS, CSSortClausesL),
	%% Have a list of lists, thus flatten:
	append(CSSortClausesL, Res4),
	%% Add clauses to make with each term its subterms domain elements,
	%% too:
	%% No - it's not needed:
	%% scan_function_symbols(CS, FSig),
 	%% dom_subterm_clauses(FSig, Res5),
	%% Finally collect the grand result:
	append([Res1, Res2, Res3, Res4], HRes),
	%% (5) Add constants for sorts:
	scan_dom_functors(HRes, DomFunctors),
	populate_sorts_clauses(DomFunctors, Res6),
	append([HRes, Res6], Res).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rr_clause((H :- B), (H :- BDom)) :-
	term_variables(H, HVars),
	term_variables(B, BVars),
	%% For *extra* variables  ...
	subtractq(HVars, BVars, HExtraVars), 
	annotate_canonical_sort_list(HExtraVars, H, HExtraVarsAnnot),
	%% and for variables occuring in body equations:
  	p_atoms(B, (=), BEquations),
  	term_variables(BEquations, BEquationsVars),
	annotate_canonical_sort_list(BEquationsVars, B, BEquationsVarsAnnot),
  	unionq(HExtraVarsAnnot, BEquationsVarsAnnot, DomVarsAnnot),
	make_doms(DomVarsAnnot, DomVarsAnnotDom),
	append(B, DomVarsAnnotDom, BDom).

% make_doms([],[]).
% make_doms([X|R],[dom(X)|DomR]) :- make_doms(R,DomR).

% Sorted version:
make_doms([], []).
make_doms([T-P/N|R], [Dom_P_N_T|DomR]) :- 
	make_dom_p_n(T-P/N, Dom_P_N_T),
	make_doms(R, DomR).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% This could be improved to sort the variables in the head only
%% if needed (when the rule fires)
sort_clause_clauses((H :- B), Res) :-
   	term_variables((H :- B), Vars),
  	clause_atoms((H :- B), Atoms),
   	annotate_all_sorts_list(Vars, Atoms, VarsAnnot),
   	make_same_sort_clauses(VarsAnnot, [], Res).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% domain_elements_axioms_from_clause((H :- B), Res)
%% Make clauses to insert domain elements 
%% (a) for the terms in the body, and
%% (b) for the terms occuring in the equations in the heads:

domain_elements_axioms_from_clause((H :- B), Res) :-
	%% (a) for the terms in the body:
	%% occuring in the body.
	maplist(domain_elements_axioms_from_body_lit, B, 
	             InsertDomElementsClauses1L),
	%% Have a list of list clauses, thus flatten:
	append(InsertDomElementsClauses1L, InsertDomElementsClauses1),
	%% (b) for the terms occuring in the equations in the heads:
	%% For non-equations we do something different, see ALTERNATIVE above.
	%% Get all equations in the head:
	p_atoms(H, (=), HEqs), 
	%% Get all terms in the head:
	terms_args_union(HEqs, HEqsTerms),
	%% Restrict to the non-variable terms:
	sublist(nonvar, HEqsTerms, HEqsFunTerms),
	%% Make the clauses that insert the domain elements.
	%% Notice the guard, B, 
	%% the body of the clause.
	%% This guard is OK, because the
	%% new domain elements have to be made only if the rule "fires",
	%% i.e. when the body is satisfied.
	annotate_all_sorts_list(HEqsFunTerms, H, HEqsFunTermsAnnot),
  	domain_elements_axioms_from_terms(HEqsFunTermsAnnot, B, 
  	                           InsertDomElementsClauses2),
	append(InsertDomElementsClauses1, InsertDomElementsClauses2, Res).


domain_elements_axioms_from_body_lit(A, Res) :-
	abstract_atom_with_positions(A, AAbst, Subst), 
	A =.. [P|_Args],
	findall(([DomHead] :- [AAbst]),
	        ( member((_X-Position)/T, Subst),
		  nonvar(T),
		  make_dom_p_n(T-P/Position, DomHead)
	        ),
		Res).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% domain_elements_axioms_from_terms(Terms, Body, Res)
%% Terms is a list of annotated terms.
%% Res is a list of clauses with
%% heads dom_P_N(T) and body Body, for each term T in Terms.

domain_elements_axioms_from_terms([], _Body, []).
domain_elements_axioms_from_terms([T-P/N|R], Body, 
	                          [([DomHead] :- Body)|RRes]) :-
	make_dom_p_n(T-P/N, DomHead),
	domain_elements_axioms_from_terms(R, Body, RRes).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

domain_elements_axioms_from_sig(PredSymbs, DomElAxioms) :-
	maplist(domain_elements_axioms_from_sig_1, PredSymbs, DomElAxiomsL),
	append(DomElAxiomsL, DomElAxioms).

domain_elements_axioms_from_sig_1(P/N, Res) :-
	make_N_variables(N, Vars),
	PBody =.. [P|Vars],
	domain_elements_axioms_from_sig_2(Vars, P/1, [PBody], Res).

domain_elements_axioms_from_sig_2([], _Pred/_Position, _Body, []).
domain_elements_axioms_from_sig_2([X|Res], Pred/Position, Body, 
	[([dom(X)] :- [Dom_P_NX]), ([Dom_P_NX] :- Body)|RRes]) :-
	make_dom_p_n(X-Pred/Position, Dom_P_NX),
	HPosition is Position+1,
	domain_elements_axioms_from_sig_2(Res, Pred/HPosition, Body, RRes).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

dom_subterm_clauses(Sig, Res) :-
	%% Sig is a list of function symbols plus arities (F/N)
	maplist(dom_subterm_clauses_1, Sig, HRes),
	%% Have a list of lists now, flatten it
	append(HRes, Res).

dom_subterm_clauses_1(F/N, Res) :-
	%% See example below to make sense of this
	make_N_variables(N, NXVars),
	FNXVars =.. [F|NXVars],
	dom_subterm_clauses_2(NXVars, F/1, [dom(FNXVars)], Res).

dom_subterm_clauses_2([], _F/_N, _Body, []).
dom_subterm_clauses_2([X|R], F/N, Body, 
	              [([dom(X)] :- [Dom_P_NX]), ([Dom_P_NX] :- Body)|RRes]) :-
	%% Clauses of Kind 1:
	make_dom_p_n(X-F/N, Dom_P_NX),
	HN is N+1,
	dom_subterm_clauses_2(R, F/HN, Body, RRes).


/*
 * Example:

Clauses of kind 1:

dom_g_1(X1) :- dom(g(X1,X2)).
dom_g_2(X2) :- dom(g(X1,X2)).

Clauses of kind 2:

dom(X) :- dom_g_1(X).
dom(X) :- dom_g_2(X).

 */


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

populate_sorts_clauses([], []).
populate_sorts_clauses([DomPred|R], [([DomPredAtom] :- [])|RRes]) :-
	concat_atom(['some_', DomPred], DomPredArg),
	DomPredAtom =.. [DomPred,DomPredArg],
	populate_sorts_clauses(R, RRes).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

abstract_atom_with_positions(A, ARes, Subst) :-
	A =.. [P|Args],
	abstract_arg_list_with_positions(Args, [], 1, ArgsRes, _NewVarsSoFar, Subst),
	ARes =.. [P|ArgsRes].

%% abstract_arg_list(Args, VarsSoFar, Position, ArgsRes, NewVarsSoFar, ArgsSubst):
abstract_arg_list_with_positions([], VarsSoFar, _Position, [], VarsSoFar, []).
abstract_arg_list_with_positions([X|R], VarsSoFar, Position, [X|RAbst], NewVarsSoFar, RSubst) :-
	%% If X is a variable that has not occured before then it may remain:
	var(X),	\+ memberq(X, VarsSoFar), !,
	%% Just go on with the rest, but remember X is present now:
	HPosition is Position+1,
	abstract_arg_list_with_positions(R, [X|VarsSoFar], HPosition, RAbst, NewVarsSoFar, RSubst).
abstract_arg_list_with_positions([T|R], VarsSoFar, Position, [X|RAbst], NewVarsSoFar, [((X-Position)/T)|RSubst]) :-
	%% In all other cases the term T has to be replaced by a fresh 
	%% variable, which we call X here.
	%% (No need to remember X, as it is fresh anyways)
	HPosition is Position+1,
	abstract_arg_list_with_positions(R, VarsSoFar, HPosition, RAbst, NewVarsSoFar, RSubst).


