%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Separation
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(separation, 
	[rr_clause_set/2,
	rr_clause/2
     ]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/data_structures').

%% (User) option: the 'abstraction_protected' procedure
%% Used in 'abstract_and_separate' below (cf. documentation there).

abstraction_protected(_) :-
	%% Don't abstract anything -- i.e. switch off separation
	nb_getval(separation_option, none).
abstraction_protected(_=_) :-
	%% Don't abstract away equations.
	nb_getval(separation_option, nonequations).
abstraction_protected(_) :- 
	%% the whole body is abstracted
	nb_getval(separation_option, full),
	!, fail. 

rr_clause_set(CS, Res) :-
	%% First separate:
	maplist(separation_clause, CS, CSSepL),
	%% Have a list of list of list clauses, thus flatten:
	append(CSSepL, CSSep),
	%% make the clauses that insert the new domain elements:
	%% (this is somewhat more precise than ALTERNATIVE below)
 	scan_predicate_symbols(CS, PSig),
 	%% but not for the dom predicate - it would result in a tautology:
 	delete(PSig, dom/1, PSigND),
	%% and not for equality, see ALTERNATIVE below
 	delete(PSigND, (=)/2, PSigNDNE),
 	domain_elements_axioms(PSigNDNE, DomElemAxioms),
 	append(CSSep, DomElemAxioms, CSSepDomElAx),
	%% Then do the range-restriction:
	maplist(rr_clause, CSSepDomElAx, CSSepDomElAxRR),
	%% add a constant if non there:
	scan_function_symbols(CS, FSig),
	( \+ member(_C/0, FSig) ->
	    Res = [([dom(a_constant)] :- [])|CSSepDomElAxRR]
	; Res = CSSepDomElAxRR
        ).
%%	Res = [([dom(a_constant)] :- [])|CSSepDomElAxRR].

%% Todo: if flattening is off then generate clauses to insert
%% with each term in the domain its subterms.

separation_clause((H :- B), ResultClauses) :-
	%% First, abstraction and separation of H :- B.
	%% This amounts to building the two "separation clauses"
	%% that resolve to the original clause H :- B.

	%% First we separate and abstract the clause body:
	abstract_and_separate(B, BAbst, BProtected, BAbstName, BAbstSubst),
	%% The abstracted body BAbst is named by BAbstName 
	%% (an atom, parametrized with at least the variables of BAbst)
	%% The clause [BAbstName] :- BAbst *is* range-restricted
	%% if "Code Alternative 2" is chosen below in
	%% abstract_and_separate. We don't assume that, though.
	SeparationClause1 = ([BAbstName] :- BAbst),

	%% Instantiate in BAbstName the abstraction variables back with 
	%% the terms: 
	apply_subst(BAbstSubst, BAbstName, BAbstNameInst),
	%% We need this to construct the second clause, 
	%% the clause with the original head H as its conclusion, 
	%% and this clause is just H :- [BAbstNameInst|BProtected].
	SeparationClause2 = (H :- [BAbstNameInst|BProtected]), 

        %% Second, domain elements generation.
	%% Generate clauses that insert new domain elements for the 
	%% non-variable terms in the body and the head of SeparationClause2.
	%% This is not needed for SeparationClause1, as by construction
	%% it does not contain a proper function term.
	%% 
	%% First, the body of SeparationClause2:
	%% Get all terms:
	terms_args_union([BAbstNameInst|BProtected], BSepCl2Terms),
	%% Restrict to the non-variable terms:
	sublist(nonvar, BSepCl2Terms, BSepCl2FunTerms),
	%% Make the clauses that insert the domain elements.
	%% Notice the guard, BAbstName, which is the head of the
	%% SeparationClause1:
	insert_dom_element_clauses(BSepCl2FunTerms, [BAbstName], 
	                           InsertDomElementsClauses1),
	%%
	%% Second, the head, but only for equations.
	%% For non-equations we do something different, see ALTERNATIVE above.
	%% Get all equations in the head:
	p_atoms(H, (=), HSepCl2Equations), 
	%% Get all terms in the head:
	terms_args_union(HSepCl2Equations, HSepCl2Terms),
	%% Restrict to the non-variable terms:
	sublist(nonvar, HSepCl2Terms, HSepCl2FunTerms),
	%% Make the clauses that insert the domain elements.
	%% Notice the guard, [BAbstNameInst|BProtected], 
	%% the body of SeparationClause2.
	%% This guard is OK, because the
	%% new domain elements have to be made only if the rule "fires",
	%% i.e. when the body is satisfied.
  	insert_dom_element_clauses(HSepCl2FunTerms, [BAbstNameInst|BProtected], 
  	                           InsertDomElementsClauses2),
	%% Finally, collect all the new clauses:
	append([[SeparationClause1, SeparationClause2],
	        InsertDomElementsClauses1,
		InsertDomElementsClauses2
	       ], HResultClauses),
	simplify(HResultClauses, ResultClauses).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% abstract_and_separate(B, BAbst, BProtected, AbstName, Subst)
%% BAbst is an abstracted version of those atoms of B (a clause body)
%% that don't classify as abstraction_protected. In BAbst, 
%% - all terms are variables
%% - no variable occurs more than once in BAbst.
%% The non-abstracted atoms of B are returned in BProtected.
%% AbstName is a new name for BAbst, i.e. a fresh atom containing
%% at least the variables in BAbst.
%% Subst is a list of pairs X/T where X is a variable and T is a term
%% that, when applied (as a substitution) to BAbstFinal yields B again.

abstract_and_separate(B, BAbst, BProtected, AbstName, Subst) :-
	%% little wrapper:
	abstract_and_separate_helper(B, [], BAbst, BProtected, Subst),
	%% Construct a new name for the abstracted body:
	gensym(abst_def_, AbstPred),

%% Code Alternative 1:

        %% This is close to Renate's suggestion about abstraction
        %% and separation.
        %% Don't use it currently , because it may results in 
	%% a non-range-restricted clause
        %% that is difficult to supply with sorts.

	%% The abstraction name depends on the variables in B ...
%%	term_variables(B, BVars),
	%% .. and the freshly introduced variables in the abstraction.
	%% The easiest way to get hold of these is just by the domain 
	%% of Subst: 
%%	subst_domain(Subst, SubstDomain),
%%	append(SubstDomain, BVars, AllVars),
	%% Notice AllVars contains *at least* the variables of BAbst,
	%% the abstracted body, but extra variables may have come in 
	%% by abstracting away terms from B, 
	%% as e.g. the term f(X) in the body of p(Y) :- f(X)=Y.  (*)
	%% In this example abstraction of the body results in the 
	%% abstracted body U=Y, Subst is [U/f(X)] and AllVars is [U,X,Y].
	%%
	%% Now we can make the new name for the abstraction:
%%	AbstName =.. [AbstPred|AllVars].
	%% By the just said AbstName may contain more (i.e. extra) variables
	%% than BAbst. Therefore the clause [AbstName] :- BAbst
	%% is possibly not range-restricted. 

%% Code Alternative 2:

        %% The abstraction depends on the variables in BAbst *only*:
	term_variables(BAbst, BAbstVars),
	AbstName =.. [AbstPred|BAbstVars].
	%% Now the clause [AbstName] :- BAbst
	%% *is* range-restricted. 
	%% This alternative seems advantageous as it leads to 
	%% fewer parameters, and the sorted case is easier to cope with:
	%% by range-restrictedness we don't have to worry what would be
	%% the right sorted domain predicates to generate.
	%% Not sure though if that all is optimal ... (Renate ?)

abstract_and_separate_helper([], _VarsSoFar, [], [], []).
abstract_and_separate_helper([A|R], VarsSoFar, [AAbst|RAbst], RProtected, Subst) :-
	\+ abstraction_protected(A), !, 
	A =.. [P|Args],
	abstract_arg_list(Args, VarsSoFar, ArgsRes, NewVarsSoFar, ArgsSubst),
	AAbst =.. [P|ArgsRes],
	abstract_and_separate_helper(R, NewVarsSoFar, RAbst, RProtected, RSubst),
	append(ArgsSubst, RSubst, Subst).
abstract_and_separate_helper([A|R], VarsSoFar, RAbst, [A|RProtected], Subst) :-
	%% In the abstraction_protected case just move A to the 'proteced' part
	abstract_and_separate_helper(R, VarsSoFar, RAbst, RProtected, Subst).

%% abstract_arg_list(Args, VarsSoFar, ArgsRes, NewVarsSoFar, ArgsSubst):
abstract_arg_list([], VarsSoFar, [], VarsSoFar, []).
abstract_arg_list([X|R], VarsSoFar, [X|RAbst], NewVarsSoFar, RSubst) :-
	%% If X is a variable that has not occured before then it may remain:
	var(X),	\+ memberq(X, VarsSoFar), !,
	%% Just go on with the rest, but remember X is present now:
	abstract_arg_list(R, [X|VarsSoFar], RAbst, NewVarsSoFar, RSubst).
abstract_arg_list([T|R], VarsSoFar, [X|RAbst], NewVarsSoFar, [(X/T)|RSubst]) :-
	%% In all other cases the term T has to be replaced by a fresh 
	%% variable, which we call X here.
	%% (No need to remember X, as it is fresh anyways)
	abstract_arg_list(R, VarsSoFar, RAbst, NewVarsSoFar, RSubst).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% insert_dom_element_clauses(Terms, Body, Res)
%% Terms is a list of terms.
%% Res is a list of clauses with
%% heads dom(T) and body Body, for each term T in Terms.

insert_dom_element_clauses([], _Body, []).
insert_dom_element_clauses([T|R], Body, [([dom(T)] :- Body)|RRes]) :-
	insert_dom_element_clauses(R, Body, RRes).

rr_clause((H :- B), (H :- BDom)) :-
	term_variables(H, HVars),
	term_variables(B, BVars),
	%% For *extra* variables  ...
	subtractq(HVars, BVars, HDomVars), 
	%% and for variables occuring in body equations:
  	p_atoms(B, (=), BEquations),
  	term_variables(BEquations, BEquationsVars),
  	unionq(HDomVars, BEquationsVars, DomVars),
	make_doms(DomVars, DomVarsDom),
	append(B, DomVarsDom, BDom).

make_doms([],[]).
make_doms([X|R],[dom(X)|DomR]) :- make_doms(R,DomR).


domain_elements_axioms(PredSymbs, DomElAxioms) :-
	maplist(domain_elements_axioms_1, PredSymbs, DomElAxiomsL),
	append(DomElAxiomsL, DomElAxioms).


domain_elements_axioms_1(P/N, Res) :-
	make_N_variables(N, Vars),
	PBody =.. [P|Vars],
	domain_elements_axioms_2(Vars, [PBody], Res).

domain_elements_axioms_2([], _Body, []).
domain_elements_axioms_2([X|Res], Body, [([dom(X)] :- Body)|RRes]) :-
	domain_elements_axioms_2(Res, Body, RRes).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
