%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Sorting
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(sorting, 
	[annotate_all_sorts/3,
	 annotate_all_sorts_list/3,
	 annotate_canonical_sort/3,
	 annotate_canonical_sort_list/3,
	 term_sort/3,
	 make_same_sort_clauses/3,
	 make_dom_p_n/2
    ]).


%% annotate_all_sorts(T, Atoms, TAnnot)
%% Annotate to the term T all sorts T has in the list of atoms Atoms.
%% Result is in TAnnot:
annotate_all_sorts(T, Atoms, TAnnot) :-
	findall(P/N, term_sort(T, Atoms, P/N), TSorts),
	annotate(T, TSorts, TAnnot).

annotate(_T, [], []).
annotate(T, [P/N|R], [T-P/N|RAnnot]) :-
	annotate(T, R, RAnnot).

%% Wrapper for lists of terms:
annotate_all_sorts_list([], _Atoms, []).
annotate_all_sorts_list([T|R], Atoms, Res) :-
	annotate_all_sorts(T, Atoms, TAnnot),
	annotate_all_sorts_list(R, Atoms, RAnnot),
	append(TAnnot, RAnnot, Res).

annotate_all_sorts_list([], _Atoms, []).
annotate_all_sorts_list([T|R], Atoms, Res) :-
	annotate_all_sorts(T, Atoms, TAnnot),
	annotate_all_sorts_list(R, Atoms, RAnnot),
	append(TAnnot, RAnnot, Res).

annotate_canonical_sort(T, Atoms, T-P/N) :-
	canonical_term_sort(T, Atoms, P/N).

%% Wrapper for lists
annotate_canonical_sort_list([], _Atoms, []).
annotate_canonical_sort_list([T|R], Atoms, [TAnnot|RAnnot]) :-
	annotate_canonical_sort(T, Atoms, TAnnot),
	annotate_canonical_sort_list(R, Atoms, RAnnot).

%% term_sort(T, A, Res)
%% Enumerate the "sorts" of the term T in an atom A, term or list of 
%% those. Res is of the the form Fun/N, meaning that T occurs as the Nth
%% argument of the function term or atom Fun (count starts at 1).
%% term_sort(X, [A|_], Res) :- term_sort(X, A, Res).
%% term_sort(X, [_|R], Res) :- term_sort(X, R, Res).
%% The above two are not needed - the code below also works for lists

% term_sort(_T, _A, dummy/dummy) :- 
%	nb_getval(sorting_option, false), !.

term_sort(T, A, Res) :- 
	A =.. [Fun| Args],
	term_sort_arg(T, Args, Fun, 1, Res).

term_sort_arg(T, [S|_], Fun, N, Fun/N) :-
	T==S.
term_sort_arg(T, [S|_], _Fun, _N, Res) :-
	compound(S), 
	term_sort(T, S, Res).
term_sort_arg(T, [_|R], Fun, N, Res) :-
	HN is N+1,
	term_sort_arg(T, R, Fun, HN, Res).


canonical_term_sort(X, A, Res) :- term_sort(X, A, Res), !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

make_same_sort_clauses(Vars, Body, Res) :-
	findall(([Dom_P1_N1_X1] :- [Dom_P2_N2_X2|Body]),
	        ( member(X1-P1/N1, Vars),
	          member(X2-P2/N2, Vars),
		  X1 == X2, 
		  \+ X1-P1/N1 = X2-P2/N2,
		  make_dom_p_n(X1-P1/N1, Dom_P1_N1_X1),
		  make_dom_p_n(X2-P2/N2, Dom_P2_N2_X2)
	        ),
		Res).

% Regarding equality, dom_=_1 and dom_=_2 are the same as the
% union of of all domains, which is just dom.
% We capture this equivalence once and for all here:
make_dom_p_n(T-(=)/_, dom(T)).
make_dom_p_n(T-P/N, Dom_P_N_T) :-
	\+ P == (=),
	concat_atom([dom,P,N],'_',H),
	Dom_P_N_T =.. [H,T].


%% Unsported:
%% make_dom_p_n(T-P/N, dom(P/N, T)).
% make_dom_p_n(T-dummy/dummy, dom(T)) :- !.

