:- module(transforms, [yarralumla_clause_set/2]).

:- use_module(misc).
:- use_module(subsumption).


%% c1([a=b] :- []).
%% c2([f(a)=b] :- [p(a,X,f(X),g(f(X,a)))]).
%% c3([] :- [p(a,X,f(X),g(f(X,a)))]).

yarralumla_clause_set(CS, CSRes) :-
	%% CS and CSY are clause sets (note: not sets of list clauses)
	EqAxioms = [
           (X=X),
	   (X=Y :- Y=X),
	   (X=Z :- X=Y, Y=Z)],
	append(EqAxioms, CS, CSEq), 
	%% We work internally with list clauses:
	maplist(to_list_clause, CSEq, CSEqL),
	maplist(yarralumla_clause, CSEqL, CSEqLYLL),
	%% Have a list of list of clauses now, flatten it:
	append(CSEqLYLL, CSEqLYL),
	%% The code leaves a lot of identical clauses, therefore:
	subsumption_filter(CSEqLYL, CSEqLYLS), 
	%% Sorts must not be empty:
	scan_dom_predicates(CSEqLYLS, DomPreds), 
	populate_sorts_clauses(DomPreds, DomPredsClauses),
	append(CSEqLYLS, DomPredsClauses, CSEqLYLSDomPredsClauses),
	%% Conversion of this clause set back to clauses:
	maplist(from_list_clause, CSEqLYLSDomPredsClauses, CSRes).


yarralumla_clause(C, CRes) :-
	%% msg('Clause ~w', [C]),

	%% Step 1: flattening
	flatten_clause(C, CFlat), %% LCFlat is a list of list clauses
	msg('Flat: ~w',[CFlat]),

	%% Step 2: range-restriction transformation
	rr_trafo_clause_set(CFlat, CRes).

rr_trafo_clause_set(CS, CSA) :-
	%% CS and CSA are sets of list clauses
	maplist(rr_trafo_clause, CS, CSAL),
	%% Have a list of list of clauses now, flatten it:
	append(CSAL, CSA).

rr_trafo_clause((H :- B), ResultClauses) :-

	%% Stage 1: Abstraction and separation of H :- B.
	%% This includes building the two clauses that resolve to the original 
	%% clause H :- B.

	%% First we separate and abstract the clause body:
	abstract_and_separate(B, BAbst, BProtected, BAbstName, BAbstSubst),
	%% The abstracted body BAbst is named by BAbstName.
	%% The clause [BAbstName] :- BAbst *is* range-restricted.
	%% if "Code Alternative 2" is chosen below in
	%% abstract_and_separate. Thus assume that for now, and obtain
	%% the first clause:
	SeparationClause1 = ([BAbstName] :- BAbst),

	%% Instantiate in BAbstName the abstraction variables back with 
	%% the terms: 
	apply_subst(BAbstSubst, BAbstName, BAbstNameInst),
	%% We need this to construct the second clause, 
	%% the clause with the original head H as its conclusion, 
	%% and this clause is just H :- [BAbstNameInst|BProtected] .
	%% However it has to be range-restricted, and we thus get
	%% the second clause:

	SeparationClause2 = (H :- [BAbstNameInst|BProtected]), 

	%% Stage 2, sorting:
	%% Unify the sorts of all occurences of the same variable.
	%% This is a non-issue without sorting, as all variables belong
	%% to the one and only sort, "dom", 

	%% Sort SeparationClause1:
   	term_variables(SeparationClause1, SepCl1Vars),
  	clause_atoms(SeparationClause1, SepCl1Atoms),
   	annotate_all_sorts_list(SepCl1Vars, SepCl1Atoms, SepCl1VarsAnnot),
	%% Notice the guard: BAbst, the body of SeparationClause1
   	make_same_sort_clauses(SepCl1VarsAnnot, [], 
 	                       SortingClauses1),

	%% Sort SeparationClause2:
%    	term_variables(SeparationClause2, SepCl2Vars),
%   	clause_atoms(SeparationClause2, SepCl2Atoms),
%    	annotate_all_sorts_list(SepCl2Vars, SepCl2Atoms, SepCl2VarsAnnot),
%    	make_same_sort_clauses(SepCl2VarsAnnot, [], 
%  	                       SortingClauses2),


	%% Sort SeparationClause2:
	%% We do it separately, for the body variables first, 
	%% and then, if the rule fires, for the whole clause 
	%% (it could be done in one sweep, but this way it's more precise):
	%% - for the body:
%  	term_variables([BAbstNameInst|BProtected], BSepCl2Vars),
%  	annotate_all_sorts_list(BSepCl2Vars, [BAbstNameInst|BProtected], 
% 	                        BSepCl2VarsAnnot),
%  	make_same_sort_clauses(BSepCl2VarsAnnot, [], SortingClauses2),
	SortingClauses2 = [],

 	%% - for the whole clause:
   	term_variables(SeparationClause2, SepCl2Vars),
	clause_atoms(SeparationClause2, SepCl2Atoms),
   	annotate_all_sorts_list(SepCl2Vars, SepCl2Atoms, SepCl2VarsAnnot),
 	%% Notice the guard is stronger now:
   	make_same_sort_clauses(SepCl2VarsAnnot, [], 
 	                       SortingClauses3),


        %% Stage 3, domain elements generation.
	%% Generate clauses that insert new domain elements for the 
	%% non-variable terms in the body and the head:
	%% First, the body.
	%% Get all terms in the body of SeparationClause2
	terms_args_union([BAbstNameInst|BProtected], BSepCl2Terms),
	%% Restrict to the non-variable terms:
	sublist(nonvar, BSepCl2Terms, BSepCl2FunTerms),
	%% Domain elements have to be made for all sorts of all terms:
	annotate_all_sorts_list(BSepCl2FunTerms, [BAbstNameInst|BProtected], 
	                        BSepCl2FunTermsAnnot),
	%% Make the clauses that insert the domain elements.
	%% Notice the guard, BAbstName:
	insert_dom_element_clauses(BSepCl2FunTermsAnnot, [BAbstName], 
	                           InsertDomElementsClauses1),

	%% Second, the head
	%% Get all terms in the head:
	terms_args_union(H, HTerms),
	%% Restrict to the non-variable terms:
	sublist(nonvar, HTerms, HFunTerms),
	%% Domain elements have to be made for all sorts of all terms:
	annotate_all_sorts_list(HFunTerms, H, HFunTermsAnnot),

	%% Make the clauses that insert the domain elements.
	%% Notice the guard, [BAbstNameInst|BProtected], which is equivalent 
	%% to the original body B. This restriction is OK, because the
	%% new domain elements have only be made if the rule "fires".
	insert_dom_element_clauses(HFunTermsAnnot, [BAbstNameInst|BProtected], 
	                        InsertDomElementsClauses2),


%% Try
%% 	term_variables(B, BVars),
%% 	annotate_all_sorts_list(BVars, B, BVarsAnnot),
%% 	make_same_sort_clauses(BVarsAnnot, [BAbstName], SortingClauses1),
%% 	%% - for the whole clause:
%%   	term_variables((H :- B), HBVars),
%%  	append(H, B, HB),
%%   	annotate_all_sorts_list(HBVars, HB, HBVarsAnnot),
%% 	%% Notice the guard is stronger now, equivalent to the original body:
%%   	make_same_sort_clauses(HBVarsAnnot, [BAbstNameInst|BProtected], 
%% 	                       SortingClauses2),


%%      Alternative: for the head and body in one sweep (less precise)
%%         SortingClauses1 = [],
%%  	term_variables((H :- B), HBVars),
%% 	append(H, B, HB),
%%  	annotate_all_sorts_list(HBVars, HB, HBVarsAnnot),
%%   	make_same_sort_clauses(HBVarsAnnot, [BAbstName], SortingClauses2),

	range_restrict_clause(SeparationClause2, SeparationClause2RR),

	%% Finally, collect all the new clauses:
	append([[SeparationClause1, SeparationClause2RR],
	        InsertDomElementsClauses1,
		InsertDomElementsClauses2,
	        SortingClauses1,
	        SortingClauses2,
	        SortingClauses3
	       ], HResultClauses),
	simplify(HResultClauses, ResultClauses).


%% make_same_sort_clauses(Vars, Body, Result)
make_same_sort_clauses(Vars, Body, Result) :-
	findall(([Dom_P1_N1_X1] :- [Dom_P2_N2_X2|Body]),
	        ( member(X1-P1/N1, Vars),
	          member(X2-P2/N2, Vars),
		  X1 == X2, 
		  \+ X1-P1/N1 = X2-P2/N2,
		  make_dom_p_n(X1-P1/N1, Dom_P1_N1_X1),
		  make_dom_p_n(X2-P2/N2, Dom_P2_N2_X2)
	        ),
		Result).


range_restrict_clause((H :- B), (H :- BDom)) :-
	term_variables(H, HVars),
	term_variables(B, BVars),
	subtractq(HVars, BVars, HExtraVars), %% For *extra* variables only
	annotate_canonical_sort_list(HExtraVars, H, HExtraVarsAnnot),
	make_doms(HExtraVarsAnnot, HExtraVarsAnnotDom),
	append(B, HExtraVarsAnnotDom, BDom).
	%% msg('Debug: ~w',[range_restrict_clause((H :- B), (H :- BDom))]).



%% WARNING: limited applicability, will apply subst just to arguments:
apply_subst(Subst, Term, Instance) :-
	Term =.. [Fun|Args],
	apply_subst_to_args(Subst, Args, ArgsInstances),
	Instance =.. [Fun|ArgsInstances].

apply_subst_to_args(_Subst, [], []).
apply_subst_to_args(Subst, [Y|R], [T|RRes]) :-
	member(X/T, Subst), X == Y, !, 
	apply_subst_to_args(Subst, R, RRes).
apply_subst_to_args(Subst, [T|R], [T|RRes]) :-
	apply_subst_to_args(Subst, R, RRes).

clause_atoms((H :- B), Atoms) :-
	append(H, B, Atoms).

%% annotate_all_sorts(T, Atoms, TAnnot)
%% Annotate to the term T all sorts T has in the list of atoms Atoms.
%% Result is in TAnnot:
annotate_all_sorts(T, Atoms, TAnnot) :-
	findall(P/N, term_sort(T, Atoms, P/N), TSorts),
	annotate(T, TSorts, TAnnot).

annotate(_T, [], []).
annotate(T, [P/N|R], [T-P/N|RAnnot]) :-
	annotate(T, R, RAnnot).

%% Wrapper for lists of terms:
annotate_all_sorts_list([], _Atoms, []).
annotate_all_sorts_list([T|R], Atoms, Res) :-
	annotate_all_sorts(T, Atoms, TAnnot),
	annotate_all_sorts_list(R, Atoms, RAnnot),
	append(TAnnot, RAnnot, Res).

annotate_all_sorts_list([], _Atoms, []).
annotate_all_sorts_list([T|R], Atoms, Res) :-
	annotate_all_sorts(T, Atoms, TAnnot),
	annotate_all_sorts_list(R, Atoms, RAnnot),
	append(TAnnot, RAnnot, Res).

annotate_canonical_sort(T, Atoms, T-P/N) :-
	canonical_term_sort(T, Atoms, P/N).

%% Wrapper for lists
annotate_canonical_sort_list([], _Atoms, []).
annotate_canonical_sort_list([T|R], Atoms, [TAnnot|RAnnot]) :-
	annotate_canonical_sort(T, Atoms, TAnnot),
	annotate_canonical_sort_list(R, Atoms, RAnnot).


%% term_sort(T, A, Res)
%% Enumerate the "sorts" of the term T in an atom A, term or list of 
%% those. Res is of the the form Fun/N, meaning that T occurs as the Nth
%% argument of the function term or atom Fun (count starts at 1).
%% term_sort(X, [A|_], Res) :- term_sort(X, A, Res).
%% term_sort(X, [_|R], Res) :- term_sort(X, R, Res).
%% The above two are not needed - the code below also works for lists

term_sort(_T, _A, dummy/dummy) :- 
	nb_getval(sorting_option, false), !.

term_sort(T, A, Res) :- 
	A =.. [Fun| Args],
	term_sort_arg(T, Args, Fun, 1, Res).

term_sort_arg(T, [S|_], Fun, N, Fun/N) :-
	T==S.
term_sort_arg(T, [S|_], _Fun, _N, Res) :-
	compound(S), 
	term_sort(T, S, Res).
term_sort_arg(T, [_|R], Fun, N, Res) :-
	HN is N+1,
	term_sort_arg(T, R, Fun, HN, Res).


canonical_term_sort(X, A, Res) :- term_sort(X, A, Res), !.

make_dom_p_n(T-dummy/dummy, dom(T)) :- !.
make_dom_p_n(T-P/N, Dom_P_N_T) :-
	concat_atom([dom,P,N],'_',H),
	Dom_P_N_T =.. [H,T].
	

terms_args_union(TermsList, ArgsUnion) :-
	maplist(args, TermsList, ArgsLists),
	append(ArgsLists, ArgsUnion).

args(Term, Args) :-
	Term =.. [_Pred|Args].

%% insert_dom_element_clauses(Terms, Body, Res)
%% Terms is a list of annotated terms.
%% Res is a list of range-restricted versions of clauses with
%% heads dom_P_N(T), for each annotated term T-P/N in Terms.
insert_dom_element_clauses([], _Body, []).
insert_dom_element_clauses([T-P/N|R], Body, [TClause|RRes]) :-
	make_dom_p_n(T-P/N, Dom_P_N_T),
	range_restrict_clause(([Dom_P_N_T] :- Body), TClause),
	insert_dom_element_clauses(R, Body, RRes).


%% Unsorted version (3 arguments):
insert_dom_element_clauses_unsorted([], _Body, []).
insert_dom_element_clauses_unsorted([T|R], Body, [TClause|RRes]) :-
	range_restrict_clause(([dom(T)] :- Body), TClause),
	insert_dom_element_clauses_unsorted(R, Body, RRes).


%% make_doms(Terms, Dom)
%% Terms is a list of annotated terms of the form T-P/N, and Dom is a list 
%% of corresponding dom atoms of the form dom_P_N(T)

make_doms([], []).
make_doms([T-P/N|R], [Dom_P_N_T|DomR]) :- 
	make_dom_p_n(T-P/N, Dom_P_N_T),
	make_doms(R, DomR).

%% Unsorted version:
%% make_doms([],[]).
%% make_doms([X|R],[dom(X)|DomR]) :- make_doms(R,DomR).

subst_var((X/_T), X).
subst_term((_X/T), T).

subst_domain(Subst, Domain) :-
	maplist(subst_var, Subst, Domain).
subst_range(Subst, Range) :-
	maplist(subst_term, Subst, Range).
subst_codomain(Subst, Codomain) :- 
	subst_range(Subst, Range),
	term_variables(Range, Codomain).

%% abstract_and_separate(B, BAbst, BProtected, AbstName, Subst)
%% BAbst is an abstracted version of those atoms of B (a clause body)
%% that don't classify as abstraction_protected. In BAbst, 
%% - all terms are variables
%% - no variable occurs more than once in BAbst.
%% The non-abstracted atoms of B are returned in BProtected.
%% AbstName is a new name for BAbst.
%% Subst is a list of pairs X/T where X is a variable and T is a term
%% that, when applied (as a substitution) to BAbstFinal yields B again.

abstract_and_separate(B, BAbst, BProtected, AbstName, Subst) :-
	%% little wrapper:
	abstract_and_separate_helper(B, [], BAbst, BProtected, Subst),
	%% Construct a new name for the abstracted body:
	gensym(abst_def_, AbstPred),

%% Code Alternative 1:

        %% This is close to Renate's suggestion about abstraction
        %% and separation.
        %% Don't use it, because it may results in a non-range-restricted clause
        %% that is difficult to supply with sorts, if "Code Alternative 1" below
        %% is chosen. Instead use slightly different variant, which will always
        %% result in a range-restricted clause.

	%% The abstraction name depends on the variables in B ...
%%	term_variables(B, BVars),
	%% .. and the freshly introduced variables in the abstraction.
	%% The easiest way to get hold of these is just by the domain 
	%% of Subst: 
%%	subst_domain(Subst, SubstDomain),
%%	append(SubstDomain, BVars, AllVars),
	%% Notice AllVars contains *at least* the variables of BAbst,
	%% the abstracted body, but extra variables may have come in 
	%% by abstracting away terms from B, 
	%% as e.g. the term f(X) in the body of p(Y) :- f(X)=Y.  (*)
	%% In this example abstraction of the body results in the 
	%% abstracted body U=Y, Subst is [U/f(X)] and AllVars is [U,X,Y].
	%%
	%% Now we can make the new name for the abstraction:
%%	AbstName =.. [AbstPred|AllVars].
	%% By the just said AbstName may contain more (i.e. extra) variables
	%% than BAbst. Therefore the clause [AbstName] :- BAbst
	%% is possibly not range-restricted. 

%% Code Alternative 2:

        %% The abstraction depends on the variables in BAbst alone:
        term_variables(BAbst, BAbstVars),
	AbstName =.. [AbstPred|BAbstVars].
	%% Now the clause [AbstName] :- BAbst
	%% *is* range-restricted. 
	%% This alternative seems advantageous as it leads to 
	%% fewer parameters, and the sorted case is easier to cope with:
	%% by range-restrictedness we don't have to worry what would be
	%% the right sorted domain predicates to generate.


abstraction_protected(_=_). %% Currently we don't abstract away equations.
%% abstraction_protected(_) :- !, fail.

abstract_and_separate_helper([], _VarsSoFar, [], [], []).
abstract_and_separate_helper([A|R], VarsSoFar, [AAbst|RAbst], RProtected, Subst) :-
	\+ abstraction_protected(A), !, 
	A =.. [P|Args],
	abstract_arg_list(Args, VarsSoFar, ArgsRes, NewVarsSoFar, ArgsSubst),
	AAbst =.. [P|ArgsRes],
	abstract_and_separate_helper(R, NewVarsSoFar, RAbst, RProtected, RSubst),
	append(ArgsSubst, RSubst, Subst).
abstract_and_separate_helper([A|R], VarsSoFar, RAbst, [A|RProtected], Subst) :-
	%% In the abstraction_protected case just move A to the 'proteced' part
	abstract_and_separate_helper(R, VarsSoFar, RAbst, RProtected, Subst).

%% abstract_arg_list(Args, VarsSoFar, ArgsRes, NewVarsSoFar, ArgsSubst):
abstract_arg_list([], VarsSoFar, [], VarsSoFar, []).
abstract_arg_list([X|R], VarsSoFar, [X|RAbst], NewVarsSoFar, RSubst) :-
	%% If X is a variable that has not occured before then it may remain:
	var(X),	\+ memberq(X, VarsSoFar), !,
	%% Just go on with the rest, but remember X is present now:
	abstract_arg_list(R, [X|VarsSoFar], RAbst, NewVarsSoFar, RSubst).
abstract_arg_list([T|R], VarsSoFar, [X|RAbst], NewVarsSoFar, [(X/T)|RSubst]) :-
	%% In all other cases the term T has to be replaced by a fresh 
	%% variable, which we call X here.
	%% (No need to remember X, as it is fresh anyways)
	abstract_arg_list(R, VarsSoFar, RAbst, NewVarsSoFar, RSubst).

flatten_clause((H :- B), [HFlat :- BRes]) :-
	nb_getval(flattening_option, standard),
	%% H :- B is a list clause.
	%% Notice the result is a singleton.
	flatten_list(H, HFlat, HDefs),
	flatten_list(B, BFlat, BDefs),
	append([BFlat, HDefs, BDefs], BRes).


flatten_clause((H :- B), Res) :-
	nb_getval(flattening_option, separation),
	%% H :- B is a list clause.
	%% Separate away the definitions for the extracted subterms
	%% Notice this is not done recursively
	%% (for lazyness reasons; also it will not be needed in practice (?)).
	flatten_list(H, HFlat, HDefs),
	flatten_list(B, BFlat, BDefs),
	%% Now separate away the HDefs and BDefs instead of integrating
	%% them into the clause body.
	%% Start with the head:
	term_variables((HFlat :- BFlat), Vars),
	term_variables(HDefs, HDefsVars),
	intersectq(Vars, HDefsVars, HDefsCommonVars),
	gensym(head_def_, HDefsSepPred),
	HDefsSepAtom =.. [HDefsSepPred|HDefsCommonVars],
	%% The same for the body:
	term_variables(BDefs, BDefsVars),
	intersectq(Vars, BDefsVars, BDefsCommonVars),
	gensym(body_def_, BDefsSepPred),
	BDefsSepAtom =.. [BDefsSepPred|BDefsCommonVars],
	%% Now some sanity check: if the separated clause is a unit clause
	%% there is no point in separating it away. 
	%% (In fact, it must be a ground clause then.)
	%% Get rid of it by unit resolution:
        HRes = [(HFlat :- [HDefsSepAtom, BDefsSepAtom | BFlat]),
               ([HDefsSepAtom] :- HDefs),
               ([BDefsSepAtom] :- BDefs)],
        simplify(HRes, Res).
		      

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% flatten(A, AFlat, ADefs)
%% Flatten an atom or term A, yielding the atom or term AFlat and a set of 
%% definitions (equations) ADefs for the extracted subterms, 
%% which are all flat, too.
flatten(N, N, []) :- 
	%% Just allow numbers in the input
	number(N), !.

flatten(X, X, []) :- 
	var(X), !.

flatten(S=T, SFlat=TFlat, STDefs) :- !,
	%% For an equation to be flat, both sides have to be flat terms
	flatten(S, SFlat, SDefs),
	flatten(T, TFlat, TDefs),
	append(SDefs, TDefs, STDefs).

flatten(A, AFlat, ADefs) :-
	%% Flatten an atom that is not an equation or 
	%% that is a (compound) term 
	A =.. [Pred|Args],
	flatten_args(Args, ArgsFlat, ADefs),
	AFlat =.. [Pred|ArgsFlat].

flatten_args([], [], []).

flatten_args([X|R], [X|RFlat], RDefs) :- 
	%% Nothing to do when flattening a variable
	var(X), !,
	flatten_args(R, RFlat, RDefs).

flatten_args([T|R], [X|RFlat], TRDefs) :- 
	%% Extract the term T, because it is not a variable
	flatten(T=X, TFlat=X, TDefs),
	flatten_args(R, RFlat, RDefs),
	append([TFlat=X| TDefs], RDefs, TRDefs).

%% lifting of flatten/3 to lists:
flatten_list([],[],[]).
flatten_list([A|R], [AFlat|RFlat], ARDefs) :-
	flatten(A, AFlat, ADefs),
	flatten_list(R, RFlat, RDefs),
	append(ADefs, RDefs, ARDefs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

scan_dom_predicates(CS, DomPreds) :-
	maplist(scan_dom_predicates_clause, CS, ClauseDomPredsLists),
	%% Have a list of lists now, flatten it:
	append(ClauseDomPredsLists, ClauseDomPreds),
	%% Eliminate duplicates:
	list_to_set(ClauseDomPreds, DomPreds).


scan_dom_predicates_clause((H :- B), DomPreds) :-
	append(H, B, HB),
	findall(Pred,
	        ( member(A, HB), 
		  A =.. [Pred|_],
		  sub_atom(Pred, 0, 3, _, 'dom')),
		DomPreds).
		 
populate_sorts_clauses([], []).
populate_sorts_clauses([DomPred|R], [([DomPredAtom] :- [])|RRes]) :-
	concat_atom(['some_', DomPred], DomPredArg),
	DomPredAtom =.. [DomPred,DomPredArg],
	populate_sorts_clauses(R, RRes).

