:- module(transforms, [yarralumla_clause_set/2]).

:- use_module(misc).


c1([a=b] :- []).
c2([f(a)=b] :- [p(a,X,f(X),g(f(X,a)))]).
c3([] :- [p(a,X,f(X),g(f(X,a)))]).

yarralumla_clause_set(CS, CSY) :-
	%% CS and CSY are clause sets (note: not sets of list clauses)
	maplist(yarralumla_clause, CS, CSYL),
	%% Have a list of list of clauses now, flatten it:
	append(CSYL, HCSY),
	%% Add constant code:
	%% Axioms of equivalence relation:
	append(HCSY,[
           dom(some_constant), %% at least one constant
           (X=X :- dom(X)),  %% Reflexivity should be domain restricted
	   (X=Y :- Y=X),
	   (X=Z :- X=Y, Y=Z)
           ], CSY).


yarralumla_clause(C, CRes) :-
	msg('Clause ~w', [C]),
	%% We work internally with list clauses:
	to_list_clause(C, LC),
	%% Step 1: flattening
	flatten_clause(LC, LCFlat), %% LCFlat is a list of list clauses
	msg('Flat: ~w',[LCFlat]),
	%% Step 2: abstraction
	abstract_clause_set(LCFlat,LCFlatAbstracted),
	%% Conversion back to clauses:
	%% msg('Abstracted: ~w',[LCFlatAbstracted]),
	maplist(from_list_clause, LCFlatAbstracted, CRes).

abstract_clause_set(CS, CSA) :-
	%% CS and CSA are sets of list clauses
	maplist(abstract_clause, CS, CSAL),
	%% Have a list of list of clauses now, flatten it:
	append(CSAL, CSA).

abstract_clause((H :- B), [([AbstName] :- BAbstFinal), 
	                   (H :- [AbstPartInst|Dom])|
		           NewDomainClauses]) :-
	%% both C and CRes are flat list clauses
	abstract_clause_body(B, BAbst, Subst),
	%% The predicate used for the definition of the abstraction:
	gensym(abst_def_, AbstPred),
	%% The abstraction depends on the variables in B ...
	term_variables(B, BVars),
	%% .. and the freshly introduced variables in Subst:
	subst_domain(Subst, SubstDomain),
	append(SubstDomain, BVars, AllVars),
	%% Notice AllVars contains at least the variables of BAbst,
	%% extra variables may come in by abstracting away terms from
	%% B, as f(X) in p(Y) :- f(X)=Y. (*)
	%% Now we can make the new name for the abstraction:
	AbstName =.. [AbstPred|AllVars],
	%% Notice that AbstName may contain more (i.e. extra) variables
	%% than BAbst. Therefore the clause [AbstName] :- BAbst
	%% is possibly not range-restricted. 
	%% Thus we make it range-restricted (cf (*) above to see why this
	%% works):
	%% subst_codomain(Subst, SubstCodomain),
	%% make_doms(SubstCodomain, BVarsFinalDom),
	%% Or, simpler, add domain restriction for the variables BVars.
	%% However, no need for the domain restriction for those variables 
	%% that occur elsewhere in BAbst:
	%% (This changes with blocking):
	term_variables(BAbst, BAbstVars),
	subtractq(BVars, BAbstVars, BVarsFinal),
	make_doms(BVarsFinal, BVarsFinalDom),
	append(BAbst, BVarsFinalDom, BAbstFinal),
	%% Instantiate in AbstName the abstraction variables back with 
	%% the terms. We need this to derive the head:
	subst_range(Subst, SubstRange),
	append(SubstRange, BVars, BodyRecoverTerms),
	AbstPartInst =.. [AbstPred|BodyRecoverTerms],
	%% Create the domain restriction for the extra head variables:
	term_variables(H, HVars),
	subtractq(HVars, BVars, ExtraVars),
	make_doms(ExtraVars, Dom),
	%% Insert new domain elements for the terms in the body.
	%% These are just the non-variable terms previous abstracted away,
	%% guarded by the name of the abstracted body.
	sublist(nonvar, SubstRange, NewDomainElementsBody),
	make_new_domain_clauses(NewDomainElementsBody, [AbstName], 
	                        NewDomainClausesBody),
	%% Insert new domain elements for the terms in the head.
	%% These are just its non-variable terms, guarded by the original 
	%% body, or, equivalently, AbstPartInst.
	terms_args_union(H, HTerms),
	sublist(nonvar, HTerms, NewDomainElementsHead),
	make_new_domain_clauses(NewDomainElementsHead, [AbstPartInst], 
	                        NewDomainClausesHead),
	append(NewDomainClausesBody, NewDomainClausesHead, NewDomainClauses).
	

terms_args_union(TermsList, ArgsUnion) :-
	maplist(args, TermsList, ArgsLists),
	append(ArgsLists, ArgsUnion).

args(Term, Args) :-
	Term =.. [_Pred|Args].

make_new_domain_clauses([], _Body, []).
make_new_domain_clauses([T|R], Body, [([dom(T)] :- Body)|RRes]) :-
	make_new_domain_clauses(R, Body, RRes).


make_doms([],[]).
make_doms([X|R],[dom(X)|DomR]) :- make_doms(R,DomR).

subst_var((X/_T), X).
subst_term((_X/T), T).

subst_domain(Subst, Domain) :-
	maplist(subst_var, Subst, Domain).
subst_range(Subst, Range) :-
	maplist(subst_term, Subst, Range).
subst_codomain(Subst, Codomain) :- 
	subst_range(Subst, Range),
	term_variables(Range, Codomain).

%% abstract_clause_body(B, BRes, Subst)
%% BRes is an abstracted version of the list of atoms B (a clause body):
%% - all terms are variables
%% - no variable occurs more than once in BRes
%% Subst is a list of pairs X/T where X is a variable and T is a term
%% that, when applied (as a substitution) to BRes yields B again.
abstract_clause_body(B, BRes, Subst) :-
	%% little wrapper:
	abstract_clause_body(B, [], BRes, Subst).

abstract_clause_body([], _VarsSoFar, [], []).
abstract_clause_body([A|R], VarsSoFar, [ARes|RRes], Subst) :-
	A =.. [P|Args],
	abstract_arg_list(Args, VarsSoFar, ArgsRes, NewVarsSoFar, ArgsSubst),
	ARes =.. [P|ArgsRes],
	abstract_clause_body(R, NewVarsSoFar, RRes, RSubst),
	append(ArgsSubst, RSubst, Subst).

%% abstract_arg_list(Args, VarsSoFar, ArgsRes, NewVarsSoFar, ArgsSubst):
abstract_arg_list([], VarsSoFar, [], VarsSoFar, []).
abstract_arg_list([X|R], VarsSoFar, [X|RRes], NewVarsSoFar, RSubst) :-
	%% If X is a variable that has not occured before then it may remain:
	var(X),	\+ memberq(X, VarsSoFar), !,
	%% Just go on with the rest, but remember X is present now:
	abstract_arg_list(R, [X|VarsSoFar], RRes, NewVarsSoFar, RSubst).
abstract_arg_list([T|R], VarsSoFar, [X|RRes], NewVarsSoFar, [(X/T)|RSubst]) :-
	%% In all other cases the term T has to be replaced by a fresh 
	%% variable, which we call X here.
	%% (No need to remember X, as it is fresh anyways)
	abstract_arg_list(R, VarsSoFar, RRes, NewVarsSoFar, RSubst).

flatten_clause((H :- B), [HFlat :- BRes]) :-
	nb_getval(flattening_option, standard),
	%% H :- B is a list clause.
	%% Notice the result is a singleton.
	flatten_list(H, HFlat, HDefs),
	flatten_list(B, BFlat, BDefs),
	append([BFlat, HDefs, BDefs], BRes).


flatten_clause((H :- B), Res) :-
	nb_getval(flattening_option, separation),
	%% H :- B is a list clause.
	%% Separate away the definitions for the extracted subterms
	%% Notice this is not done recursively
	%% (for lazyness reasons; also it will not be needed in practice (?)).
	flatten_list(H, HFlat, HDefs),
	flatten_list(B, BFlat, BDefs),
	%% Now separate away the HDefs and BDefs instead of integrating
	%% them into the clause body.
	%% Start with the head:
	term_variables((HFlat :- BFlat), Vars),
	term_variables(HDefs, HDefsVars),
	intersectq(Vars, HDefsVars, HDefsCommonVars),
	gensym(head_def_, HDefsSepPred),
	HDefsSepAtom =.. [HDefsSepPred|HDefsCommonVars],
	%% The same for the body:
	term_variables(BDefs, BDefsVars),
	intersectq(Vars, BDefsVars, BDefsCommonVars),
	gensym(body_def_, BDefsSepPred),
	BDefsSepAtom =.. [BDefsSepPred|BDefsCommonVars],
	%% Now some sanity check: if the separated clause is a unit clause
	%% there is no point in separating it away. 
	%% (In fact, it must be a ground clause then.)
	%% Get rid of it by unit resolution:
	( (HDefs == [], BDefs == []) -> 
	    Res = [(HFlat :- BFlat)]
	; HDefs == [] -> %% That is, BDefs is not empty
	    Res = [(HFlat :- [BDefsSepAtom | BFlat]),
                          ([BDefsSepAtom] :- BDefs)]
	; BDefs == [] -> %% That is, HDefs is not empty
	    Res = [(HFlat :- [HDefsSepAtom | BFlat]),
                          ([HDefsSepAtom] :- HDefs)]
	; %% Both HDefs and BDefs are not empty
          Res = [(HFlat :- [HDefsSepAtom, BDefsSepAtom | BFlat]),
                          ([HDefsSepAtom] :- HDefs),
                          ([BDefsSepAtom] :- BDefs)]
	).
		      

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% flatten(A, AFlat, ADefs)
%% Flatten an atom or term A, yielding the atom or term AFlat and a set of 
%% definitions (equations) ADefs for the extracted subterms, 
%% which are all flat, too.
flatten(N, N, []) :- 
	%% Just allow numbers in the input
	number(N), !.

flatten(X, X, []) :- 
	var(X), !.

flatten(S=T, SFlat=TFlat, STDefs) :- !,
	%% For an equation to be flat, both sides have to be flat terms
	flatten(S, SFlat, SDefs),
	flatten(T, TFlat, TDefs),
	append(SDefs, TDefs, STDefs).

flatten(A, AFlat, ADefs) :-
	%% Flatten an atom that is not an equation or 
	%% that is a (compound) term 
	A =.. [Pred|Args],
	flatten_args(Args, ArgsFlat, ADefs),
	AFlat =.. [Pred|ArgsFlat].

flatten_args([], [], []).

flatten_args([X|R], [X|RFlat], RDefs) :- 
	%% Nothing to do when flattening a variable
	var(X), !,
	flatten_args(R, RFlat, RDefs).

flatten_args([T|R], [X|RFlat], TRDefs) :- 
	%% Extract the term T, because it is not a variable
	flatten(T=X, TFlat=X, TDefs),
	flatten_args(R, RFlat, RDefs),
	append([TFlat=X| TDefs], RDefs, TRDefs).

%% lifting of flatten/3 to lists:
flatten_list([],[],[]).
flatten_list([A|R], [AFlat|RFlat], ARDefs) :-
	flatten(A, AFlat, ADefs),
	flatten_list(R, RFlat, RDefs),
	append(ADefs, RDefs, ARDefs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



