:- module(transforms, [yarralumla_clause_set/2]).

:- use_module('../yarralumla/misc').
:- use_module('../yarralumla/flattening').
:- use_module('../yarralumla/shifting').
:- use_module('../yarralumla/range_restriction').
:- use_module('../yarralumla/data_structures').
%% :- use_module(sorting).
:- use_module('../yarralumla/subsumption').



%% Main procedure:
yarralumla_clause_set(CS, CSRes) :-
	%% CS and CSRes are lists of clauses (note: not lists of list clauses)
	%% Add equivalence relation axioms:
	( nb_getval(eqaxioms_option, true) ->
	    EqAxioms = [(X=X),
			(X=Y :- Y=X),
			(X=Z :- X=Y, Y=Z)],
	    append(EqAxioms, CS, CSEq)
	; CSEq = CS
        ),
	%% We work internally with list clauses
	%% (cf. datastructures.pl for a description of list clauses):
	maplist(to_list_clause, CSEq, CS1),
	%% Add predicate substitution axioms if necessary:
	( ( nb_getval(flattentoplevel_option, false),
	    nb_getval(eqaxioms_option, true) ) ->
	    scan_predicate_symbols(CS1, HOrigPredSymbs),
	    list_to_set(HOrigPredSymbs, OrigPredSymbs),
	    delete(OrigPredSymbs, (=)/2, OrigPredSymbsNE),
	    predicate_substitution_axioms(OrigPredSymbsNE, SubstAxioms),
	    append(CS1, SubstAxioms, CS2)
	; CS2 = CS1),
	%% Initialization of (ugly) global variables:
	%% set_create(predicate_symbols),
	set_create(not_predicate_symbols),
	%% Apply the main transformation:
	%% (A little unsystematic - better apply individual transformations
	%%  to clause sets. Unfortunately, this is not quite possible.)
	maplist(yarralumla_clause, CS2, CS3L),
	%% Have a list of list of clauses now, flatten it:
	append(CS3L, CS3),
	%%
	%% Make the clauses to generate the new domain elements:
	%% that is, generate clauses of the form dom(X) :- p(X).
	%% no - not sufficient, have also to make domain elements
	%% for function terms in the body
%%  	( nb_getval(rangerestriction_option, true) ->
%%  	    scan_predicate_symbols(CS3, HPredSymbs),
%%  	    list_to_set(HPredSymbs, PredSymbs),
%%  	    delete(PredSymbs, dom/1, PredSymbsND),
%%  	    domain_elements_axioms(PredSymbsND, DomElemAxioms),
%%  	    append(CS3, DomElemAxioms, CS4)
%%  	; CS4 = CS3),
        CS4 = CS3,

	%% Sorts must not be empty:
	%% Non-issue for unsorted case:
%%      scan_dom_predicates(CS2S, DomPreds), 
%%      populate_sorts_clauses(DomPreds, DomPredsClauses),
	%% Without sorts just add a dummy constant:
	DomPredsClauses = [[dom(some_constant)] :- []],
	append(CS4, DomPredsClauses, CS5),

	%% Add consistency axioms for predicate symbols of shifted atoms
	( nb_getval(shifting_option, true) ->
	    set_members(not_predicate_symbols, PredicateSymbols),
	    consistency_clauses(PredicateSymbols, ConsistencyClauses),
            append(CS5, ConsistencyClauses, CS6)
	; CS6 = CS5
        ),

	( nb_getval(blocking_option, true) ->
	    %% Make the clauses to speculate equality with subterms:
	    scan_function_symbols(CS6, HFunSymbs),
	    list_to_set(HFunSymbs, FunSymbs),
	    subterm_clauses(FunSymbs, SubtermClauses),
	    append(CS6, SubtermClauses, CS7H),
	    %% Make the clauses to rewrite with non-domain 
	    %% element left hand sides:
	    ( nb_getval(eqaxioms_option, true) ->
		( nb_getval(flattentoplevel_option, true) ->
		  %% Proper function terms can only occur in (non-)equations then
		  %% And we need to consider proper function terms only
		  %% (Constants cannot be equated with a subterm)
		  RewriteSig = [(=)/2, ('not_='/2)]
	        ; scan_predicate_symbols(CS6, RewriteSigH1),
		  delete(RewriteSigH1, dom/1, RewriteSigH2),
		  delete(RewriteSigH2, dom_candidate/1, H3RewriteSig),
		  list_to_set(H3RewriteSig, RewriteSig)
	        ),
		predicate_rewrite_axioms(RewriteSig, RewriteAxioms),
		append(CS7H, RewriteAxioms, CS7)
	    ; %% eqaxioms_option is false
	      CS7 = CS7H
	    )
	; %% blocking_option is false
	%% append(CS6, [[dom(X)]:-[dom_candidate(X)]], CS7)
	CS7 = CS6
        ),

	%% The code leaves a lot of identical clauses, therefore:
	%% subsumption_filter(CS7, CS7S), 
	CS7 = CS7S, 
	%% Currrently disabled, as subsumption deletion is
	%% important in conjunction with sorting, but not so without
	%% sorting, and it is very costly:

	%%
	%% Conversion of this clause set back to clauses:
	maplist(from_list_clause, CS7S, CSRes).

%% yarralumla_clause(C, CRes)
%% All the transformations, applied to the list clause C, giving 
%% the result CRes
yarralumla_clause(C, CRes) :-
	%% msg('Clause ~w', [C]),
	%% Step 1: flattening
	flatten_clause(C, CFlat), %% LCFlat is a list of list clauses
	%% msg('Flat: ~w',[CFlat]),
	%% Step 2: shifting
	( nb_getval(shifting_option, true) ->
	    shift_clause_set(CFlat, CFlatShift)
	; CFlatShift = CFlat
        ),
	%% Step 3: range-restriction transformation
	rr_clause_set(CFlatShift, CRes).
        %% ToDo: sorting


predicate_substitution_axioms(PredSymbs, SubstAxioms) :-
	maplist(predicate_substitution_axiom, PredSymbs, SubstAxioms).

predicate_substitution_axiom(P/N, ([PHead] :- [PBody|Equations])) :-
	make_N_variables(N, Vars1),
	make_N_variables(N, Vars2),
	make_equations(Vars1, Vars2, Equations),
	PHead =.. [P|Vars2],
	PBody =.. [P|Vars1].

predicate_rewrite_axioms(PredSymbs, RewriteAxioms) :-
	maplist(predicate_rewrite_axioms_, PredSymbs, RewriteAxiomsL),
	%% Have a list of list, thus flatten:
	append(RewriteAxiomsL, RewriteAxioms).

predicate_rewrite_axioms_(P/N, Res) :-
	make_N_variables(N, NVars),
	findall(([H] :- [B1, X = XPrime, equals_some_subterm_(X, _)]),
	        ( member(X, NVars),
		  replaceq(NVars, X, XPrime, NVarsPrime),
		  H =.. [P|NVarsPrime],
		  B1 =.. [P|NVars]
	        ),
		Res).

consistency_clauses([], []).
consistency_clauses([P/N|R], [([] :- [B1,B2])|RRes]) :-
	make_N_variables(N, NVars),
	B1 =.. [P|NVars],
	concat_atom(['not_',P], NotP),
	B2 =.. [NotP|NVars],
	consistency_clauses(R, RRes).



/*
 * Unsorted case: not needed

scan_dom_predicates(CS, DomPreds) :-
	maplist(scan_dom_predicates_clause, CS, ClauseDomPredsLists),
	%% Have a list of lists now, flatten it:
	append(ClauseDomPredsLists, ClauseDomPreds),
	%% Eliminate duplicates:
	list_to_set(ClauseDomPreds, DomPreds).


scan_dom_predicates_clause((H :- B), DomPreds) :-
	append(H, B, HB),
	findall(Pred,
	        ( member(A, HB), 
		  A =.. [Pred|_],
		  sub_atom(Pred, 0, 3, _, 'dom')),
		DomPreds).
		 
populate_sorts_clauses([], []).
populate_sorts_clauses([DomPred|R], [([DomPredAtom] :- [])|RRes]) :-
	concat_atom(['some_', DomPred], DomPredArg),
	DomPredAtom =.. [DomPred,DomPredArg],
	populate_sorts_clauses(R, RRes).

 */


